@@ -30,7 +30,7 @@ def get_usernames(user_ids, token=None):
     if token is None:
         raise ValueError('missing token')
     if len(user_ids):
-        users_request = requests.get('https://' + AUTH_URL + '/api/v1/user?id={}'.format(','.join(map(str, user_ids))),
+        users_request = requests.get(f"https://{AUTH_URL}/api/v1/user?id={','.join(map(str, user_ids))}",
             headers={'authorization': 'Bearer ' + token},
             timeout=5)
 
@@ -46,7 +46,7 @@ def get_usernames(user_ids, token=None):
 
 @cachetools.func.ttl_cache(maxsize=CACHE_MAXSIZE, ttl=CACHE_TTL)
 def user_cache_http(token):
-    user_request = requests.get('https://' + AUTH_URL + '/api/v1/user/cache', headers={'authorization': 'Bearer ' + token})
+    user_request = requests.get(f"https://{AUTH_URL}/api/v1/user/cache", headers={'authorization': 'Bearer ' + token})
     if user_request.status_code == 200:
         return user_request.json()
 
@@ -63,77 +63,116 @@ def is_root_public(table_id, root_id, token):
     if root_id is None:
         return False
 
-    cip_url = 'https://{}/api/v1/table/{}/root/{}/is_public'.format(AUTH_URL, table_id, root_id)
+    url = f"https://{AUTH_URL}/api/v1/table/{table_id}/root/{root_id}/is_public"
 
-    user_request = requests.get(cip_url, headers={'authorization': 'Bearer ' + token}, timeout=5)
-    if user_request.status_code == 200:
-        return user_request.json()
+    req = requests.get(url, headers={'authorization': 'Bearer ' + token}, timeout=5)
+
+    if req.status_code == 200:
+        return req.json()
     else:
         raise RuntimeError('is_root_public request failed')
 
-def auth_required(f):
-    @wraps(f)
-    def decorated_function(*args, **kwargs):
-        if flask.request.method == 'OPTIONS':
-            return f(*args, **kwargs)
+@cachetools.func.ttl_cache(maxsize=CACHE_MAXSIZE, ttl=CACHE_TTL)
+def table_has_public(table_id, token):
+    url = f"https://{AUTH_URL}/api/v1/table/{table_id}/has_public"
 
-        if hasattr(flask.g, 'auth_token'):
-            # if authorization header has already been parsed, don't need to re-parse
-            # this allows auth_required to be an optional decorator if auth_requires_role is also used
-            return f(*args, **kwargs)
+    req = requests.get(url, headers={'authorization': 'Bearer ' + token}, timeout=5)
+    if req.status_code == 200:
+        return req.json()
+    else:
+        raise RuntimeError('has_public request failed')
 
-        token = None
-        cookie_name = TOKEN_NAME
+def auth_required(func=None, *, required_permission=None, public_table_key=None, public_node_key=None, service_token=None):
+    def decorator(f):
+        @wraps(f)
+        def decorated_function(*args, **kwargs):
+            if flask.request.method == 'OPTIONS':
+                return f(*args, **kwargs)
 
-        auth_header = flask.request.headers.get('authorization')
-        xrw_header = flask.request.headers.get('X-Requested-With')
+            if hasattr(flask.g, 'auth_token'):
+                # if authorization header has already been parsed, don't need to re-parse
+                # this allows auth_required to be an optional decorator if auth_requires_role is also used
+                return f(*args, **kwargs)
 
-        programmatic_access = xrw_header or auth_header or flask.request.environ.get('HTTP_ORIGIN')
+            token = None
+            cookie_name = TOKEN_NAME
 
-        AUTHORIZE_URI = 'https://' + STICKY_AUTH_URL + '/api/v1/authorize'
+            auth_header = flask.request.headers.get('authorization')
+            xrw_header = flask.request.headers.get('X-Requested-With')
 
-        query_param_token = flask.request.args.get(TOKEN_NAME)
+            programmatic_access = xrw_header or auth_header or flask.request.environ.get('HTTP_ORIGIN')
 
-        if not query_param_token:
-            # deprecated
-            query_param_token = flask.request.args.get('token')
+            AUTHORIZE_URI = 'https://' + STICKY_AUTH_URL + '/api/v1/authorize'
 
-        if programmatic_access:
-            if query_param_token:
-                token = query_param_token
-            else:
-                if not auth_header:
-                    resp = flask.Response("Unauthorized", 401)
-                    resp.headers['WWW-Authenticate'] = 'Bearer realm="' + AUTHORIZE_URI + '"'
-                    return resp
-                elif not auth_header.startswith('Bearer '):
-                    resp = flask.Response("Invalid Request", 400)
-                    resp.headers['WWW-Authenticate'] = 'Bearer realm="' + AUTHORIZE_URI + '", error="invalid_request", error_description="Header must begin with \'Bearer\'"'
+            query_param_token = flask.request.args.get(TOKEN_NAME)
+
+            if not query_param_token:
+                # deprecated
+                query_param_token = flask.request.args.get('token')
+
+            flask.g.public_access_cache = None
+            def lazy_check_public_access():
+                if flask.g.public_access_cache is None:
+                    if service_token and required_permission is not 'edit':
+                        if public_node_key is not None:
+                            flask.g.public_access_cache = is_root_public(kwargs.get(public_table_key), kwargs.get(public_node_key), service_token)
+                        elif public_table_key is not None:
+                            flask.g.public_access_cache = table_has_public(kwargs.get(public_table_key), service_token)
+                        else:
+                            flask.g.public_access_cache = False
+                    else:
+                        flask.g.public_access_cache = False
+
+                return flask.g.public_access_cache
+
+            flask.g.public_access = lazy_check_public_access
+
+            if programmatic_access:
+                if query_param_token:
+                    token = query_param_token
+                else:
+                    if not auth_header:
+                        if not flask.g.public_access():
+                            resp = flask.Response("Unauthorized", 401)
+                            resp.headers['WWW-Authenticate'] = 'Bearer realm="' + AUTHORIZE_URI + '"'
+                            return resp
+                    elif not auth_header.startswith('Bearer '):
+                        resp = flask.Response("Invalid Request", 400)
+                        resp.headers['WWW-Authenticate'] = 'Bearer realm="' + AUTHORIZE_URI + '", error="invalid_request", error_description="Header must begin with \'Bearer\'"'
+                        return resp
+                    else:
+                        token = auth_header.split(' ')[1] # remove schema
+            else: # direct browser access, or a non-browser request missing auth header (user error) TODO: check user agent to deliver 401 in this case
+                if query_param_token:
+                    resp = flask.make_response(flask.redirect(furl(flask.request.url).remove([TOKEN_NAME, 'token']).url, code=302))
+                    resp.set_cookie(cookie_name, query_param_token, secure=True, httponly=True)
                     return resp
 
-                token = auth_header.split(' ')[1] # remove schema
-        else: # direct browser access, or a non-browser request missing auth header (user error) TODO: check user agent to deliver 401 in this case
-            if query_param_token:
-                resp = flask.make_response(flask.redirect(furl(flask.request.url).remove([TOKEN_NAME, 'token']).url, code=302))
-                resp.set_cookie(cookie_name, query_param_token, secure=True, httponly=True)
+                token = flask.request.cookies.get(cookie_name)
+
+            cached_user_data = get_user_cache(token) if token else None
+
+            if cached_user_data:
+                flask.g.auth_user = cached_user_data
+                flask.g.auth_token = token
+                return f(*args, **kwargs)
+            elif not programmatic_access and not flask.g.public_access():
+                return flask.redirect(AUTHORIZE_URI + '?redirect=' + quote(flask.request.url), code=302)
+            elif not flask.g.public_access():
+                resp = flask.Response("Invalid/Expired Token", 401)
+                resp.headers['WWW-Authenticate'] = 'Bearer realm="' + AUTHORIZE_URI + '", error="invalid_token", error_description="Invalid/Expired Token"'
                 return resp
+            else:
+                flask.g.auth_user = {'id': 0, 'service_account': False, 'name': '', 'email': '', 'admin': False, 'groups': [], 'permissions': {}}
+                flask.g.auth_token = None
+                return f(*args, **kwargs)
 
-            token = flask.request.cookies.get(cookie_name)
-
-        cached_user_data = get_user_cache(token) if token else None
-
-        if cached_user_data:
-            flask.g.auth_user = cached_user_data
-            flask.g.auth_token = token
-            return f(*args, **kwargs)
-        elif not programmatic_access:
-            return flask.redirect(AUTHORIZE_URI + '?redirect=' + quote(flask.request.url), code=302)
-        else:
-            resp = flask.Response("Invalid/Expired Token", 401)
-            resp.headers['WWW-Authenticate'] = 'Bearer realm="' + AUTHORIZE_URI + '", error="invalid_token", error_description="Invalid/Expired Token"'
-            return resp
-    return decorated_function
+        return decorated_function
 
+    if func:
+        return decorator(func)
+    else:
+        return decorator
 
 def auth_requires_admin(f):
     @wraps(f)
@@ -150,11 +189,10 @@ def auth_requires_admin(f):
 
     return decorated_function
 
-
-def auth_requires_permission(required_permission, node_key='node_id'):
+def auth_requires_permission(required_permission, public_table_key=None, public_node_key=None, service_token=None):
     def decorator(f):
         @wraps(f)
-        @auth_required
+        @auth_required(required_permission=required_permission, public_table_key=public_table_key, public_node_key=public_node_key, service_token=service_token)
         def decorated_function(table_id, *args, **kwargs):
             if flask.request.method == 'OPTIONS':
                 return f(*args, **{**kwargs, **{'table_id': table_id}})
@@ -188,7 +226,7 @@ def auth_requires_permission(required_permission, node_key='node_id'):
                 level_for_dataset = flask.g.auth_user['permissions'].get(dataset, 0)
                 has_permission = level_for_dataset >= required_level
 
-                if has_permission or (required_level < 2 and is_root_public(table_id, kwargs.get(node_key), flask.g.auth_token)):
+                if has_permission or flask.g.public_access(): # public_access won't be true for edit requests
                     return f(*args, **{**kwargs, **{'table_id': table_id}})
                 else:
                     resp = flask.Response("Missing permission: {0} for dataset {1}".format(required_permission, dataset), 403)
